/*
 * Module de pilotage de LED DomoTV (kernel 2.2)
 *
 * Auteur: Pierre Ficheux (pierre@alienor.fr)
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/uaccess.h>

static int base_addr = 0;
static struct miscdevice ledctl_dev;

#define LED_RED               0xFD
#define LED_ORANGE            0xFB
#define LED_OFF               0xFF

/* Lecture adresse de base */
static int read_base_addr (void)
{
  int addr;

  unsigned int msb, lsb;

  outb (0x07, 0x2E);
  outb (0x07, 0x2F);
  outb (0x60, 0x2E);
  msb = inb (0x2F);
  outb (0x61, 0x2E);
  lsb = inb (0x2F);
  addr = (msb << 8) + lsb;
  outb (0x0f, addr+1);

  return  addr;
}


/* =================================== open and close */

static int ledctl_open (struct inode *inode, struct file *filp)
{
    printk("<1>ledctl: open\n");

    MOD_INC_USE_COUNT;
    base_addr = read_base_addr ();
    return 0;
}

static int ledctl_close (struct inode *inode, struct file *filp)
{
    MOD_DEC_USE_COUNT;
}

/* =================================== read and write */

/*
 * The read() method just returns the current status of the leds.
 * Try "while true; do dd bs=1 if=/dev/ledctl count=1; sleep 1; done"
 */
static ssize_t ledctl_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
    unsigned char led_value = inb (base_addr), led_color;

    switch (led_value) {

    case LED_RED: 
      led_color = 'R';
      break;

    case LED_ORANGE: 
      led_color = 'O';
      break;

    case LED_OFF: 
      led_color = 'E';
      break;

    default:
      led_color = '?';
      break;
    }
      
    put_user (led_color, buf);
    return count;
}

/*
 * Write understands a few textual commands
 */ 
static ssize_t ledctl_write (struct file *file, const char *buf, size_t count, loff_t *ppos)
{
  unsigned char cmd;

  if (get_user (cmd, buf++))
    return -EFAULT;

  switch (cmd) {
  case 'R':
    outb (LED_RED, base_addr);
    break;

  case 'O':
    outb (LED_ORANGE, base_addr);
    break;

  case 'E':
    outb (LED_OFF, base_addr);
    break;
    
  default:
    break;
  }
  
  return count;

}




/* =================================== all the operations */

struct file_operations ledctl_fops = {
    NULL,
    ledctl_read,
    ledctl_write,
    NULL,          /* ledctl_readdir */
    NULL,          /* ledctl_poll */
    NULL,          /* ledctl_ioctl */
    NULL,          /* ledctl_mmap */
    ledctl_open,
    NULL,
    ledctl_close,
    NULL,          /* ledctl_fsync */
    NULL,          /* ledctl_fasync */
};


/* =================================== init and cleanup */

int init_module(void)
{
    int retval;

    printk("<1>ledctl: init_module\n");

    ledctl_dev.minor = MISC_DYNAMIC_MINOR;
    ledctl_dev.name = "ledctl";
    ledctl_dev.fops = &ledctl_fops;
    retval = misc_register(&ledctl_dev);
    if (retval) return retval;
    return 0;
}

void cleanup_module(void)
{
    printk("<1>ledctl: cleanup_module\n");
    misc_deregister(&ledctl_dev);
}

